indexing

	description:

		"Move elements from source to drain%
		%Source is required, but drain is optional%
		%Pump stops automatically at end of input.%
		%Pump supports notion of error. As soon as source or drain report%
		%that they're not ok, the pump stops. When pump is restarted, it%
		%is assumed that error correction has taken place, so errors are%
		%always reset when pump starts."

	library: "Gobo Eiffel Pipework Library"
	copyright: "Copyright (c) 2001, Simon Parker and others"
	license: "Eiffel Forum License v1 (see forum.txt)"
	date: "$Date: $"
	revision: "$Revision: $"


class

	PW_PIPE_PUMP [ELEMENT]

inherit

	PW_LAST_RESULT

creation

	make,
	connect_source,
	connect

feature -- Initialization

	make is
			-- Create a new, disconnected pump.
		do
		ensure
			idle: is_idle
			disconnected: not source_is_connected and not drain_is_connected
		end

feature -- Status

	is_idle: BOOLEAN is
			-- Is the pump idle?
		do
			Result := state = idle_state
		end

	is_running: BOOLEAN is
			-- Is the pump ready for a cycle?
		do
			Result := state = running_state
		end

	is_stopping: BOOLEAN is
			-- Should the pump stop after this cycle?
		do
			Result := state = stopping_state
		end

	is_connected: BOOLEAN is
			-- Are both source and drain connected?
		do
			Result := source_is_connected and drain_is_connected
		ensure
			both_connected: Result implies source_is_connected and drain_is_connected
		end

	source_is_connected: BOOLEAN is
			-- Is a source of elements connected to the source?
		do
			Result := source /= Void
		end

	drain_is_connected: BOOLEAN is
			-- Is there a destination for elements?
		do
			Result := drain /= Void
		end

feature -- Commands

	begin is
			-- Prepare for `cycle'.
		require
			idle: is_idle
			source_connected: source_is_connected
			source_rewindable: source.is_rewindable
		do
			prime
		ensure
			primed: is_running or is_stopping
		end

	cycle is
			-- Move one object through.
		require
			primed: is_running
			no_errors: is_ok
			source_reported_no_errors: source.is_ok
			source_not_at_end: not source.end_of_input
		do
			blow
			if is_ok then
				suck
			end
		ensure
			running_or_stopping: is_running or is_stopping
		end

	finish is
			-- Finish `cycling' by becoming idle as gracefully as possible.
		do
			if not is_stopping then
				stop
			end
			if drain.is_ok then
				flush
			else
				state := idle_state
			end
		ensure
			stopped: is_idle
		end

	run is
			-- Operate until stopped.
			-- Source required, drain optional.
		require
			idle: is_idle
			source_connected: source_is_connected
			source_rewindable: source.is_rewindable
		do
			from
				prime
			invariant
				not is_idle
			until
				is_stopping
			loop
				cycle
			end
			if is_ok and then drain.is_ok then
				flush
			end
			state := idle_state
		ensure
			idle: is_idle
		end

	stop is
			-- Stop gracefully at the end of this cycle.
		require
			is_running or is_stopping
		do
			state := stopping_state
		ensure
			stopping: is_stopping
		end

feature -- Connection management

	source: PW_READER [ELEMENT]
			-- Inlet pipe, source of elements.
			-- Required for pump operation.

	drain: PW_WRITER [ELEMENT]
			-- Outlet pipe, drain for elements.
			-- Optional.

	disconnect is
			-- Disconnect both source and drain.
		require
			idle: is_idle
		do
			source := Void
			drain := Void
		ensure
			source_disconnected: not source_is_connected
			drain_disconnected: not drain_is_connected
		end

	connect (a_source: like source; a_drain: like drain) is
			-- Connect source and drain pipes.
		require
			idle: is_idle
			source_given: a_source /= Void
			drain_given: a_drain /= Void
		do
			source := a_source
			drain := a_drain
		ensure
			connected: is_connected
			source_replaced: source = a_source
			drain_replaced: drain = a_drain
		end

	connect_source (p_source: like source) is
			-- Obtain elements from p_source in future.
		require
			idle: is_idle
			source_given: p_source /= Void
		do
			source := p_source
		ensure
			source_connected: source_is_connected
			source_replaced: source = p_source
		end


	disconnect_source is
			-- Disconnect current source pipe, if any.
		require
			idle: is_idle
		do
			source := Void
		ensure
			source_disconnected: not source_is_connected
		end


	connect_drain (p_drain: like drain) is
			-- Use p_drain for output from now on.
		require
			idle: is_idle
			drain_given: p_drain /= Void
		do
			drain := p_drain
		ensure
			drain_connected: drain_is_connected
			drain_replaced: drain = p_drain
		end

	disconnect_drain is
			-- Stop using this drain.
		require
			idle: is_idle
		do
			drain := Void
		ensure
			drain_disconnected: not drain_is_connected
		end


feature {NONE} -- Internal status

	state: INTEGER
			-- Pump state.

	idle_state: INTEGER is 0
			-- Initial value for state.

	running_state,
	stopping_state: INTEGER is unique
			-- Values for state.

feature {NONE} -- Internal operations

	prime is
			-- Prepare for first stroke. Errors in `source' and `drain'
			-- are reset.
		require
			idle: is_idle
			source_connected: source_is_connected
			source_rewindable: source.is_rewindable
		do
			source.clear_error
			drain.clear_error
			source.rewind
			if not source.end_of_input then
				clear_error
				state := running_state
				suck
			else
				if not source.is_ok then
					diagnostic_code := source.diagnostic_code
					diagnostic_message := source.diagnostic_message
				end
				state := stopping_state
			end
		ensure
			error_copied: source.is_ok = is_ok
			primed: is_running or is_stopping
			only_run_if_source_ok: is_running implies source.is_ok
		end

	blow is
			-- Write last sucked element to drain,
			-- if drain is connected.
		require
			running: is_running
			source_reported_no_errors: source.is_ok
			source_not_at_end: not source.end_of_input
			drain_ok: drain.is_ok
		do
			if drain_is_connected then
				drain.put (source.last_item)
				if not drain.is_ok then
					stop
					diagnostic_code := drain.diagnostic_code
					diagnostic_message := drain.diagnostic_message
				end
			end
		ensure
			error_copied: drain.is_ok = is_ok
			only_run_if_everything_ok: is_running implies is_ok and drain.is_ok
			stopped_on_error: not drain.is_ok implies is_stopping
		end

	suck is
			-- Attempt to read next element.
		require
			running: is_running
			not_at_end: not source.end_of_input
			source_ok: source.is_ok
		do
			source.read
			if not source.is_ok then
				stop
				diagnostic_code := source.diagnostic_code
				diagnostic_message := source.diagnostic_message
			elseif source.end_of_input then
				stop
			end
		ensure
			error_copied: source.is_ok = is_ok
			stop_at_end: source.end_of_input implies is_stopping
			stop_at_error: not source.is_ok implies is_stopping
		end

	flush is
			-- Flush drain pipe before stopping,
			-- if drain is connected.
		require
			stopping: is_stopping
			drain_ok: drain.is_ok
		do
			if drain_is_connected then
				drain.flush
				if not drain.is_ok then
					diagnostic_code := drain.diagnostic_code
					diagnostic_message := drain.diagnostic_message
				end
			end
			state := idle_state
		ensure
			error_copied: drain.is_ok = is_ok
			stopped: is_idle
		end

invariant

	source_connected_if_running: is_running implies source_is_connected
	stopping_at_end: (source_is_connected and then source.end_of_input) implies (is_stopping or is_idle)
	known_state: is_idle or is_running or is_stopping
	stopped_on_error: is_running implies is_ok

end
